@@ -8,18 +8,20 @@ use JSON::Validator::Error;
 use Mojo::File 'path';
 use Mojo::JSON::Pointer;
 use Mojo::JSON;
+use Mojo::Loader;
 use Mojo::URL;
-use Mojo::Util 'deprecated';
-use Scalar::Util;
+use Mojo::Util 'url_unescape';
+use Scalar::Util qw(blessed refaddr);
 use Time::Local ();
 
+use constant DEBUG           => $ENV{JSON_VALIDATOR_DEBUG}           || 0;
+use constant RECURSION_LIMIT => $ENV{JSON_VALIDATOR_RECURSION_LIMIT} || 100;
+use constant SPECIFICATION_URL => 'http://json-schema.org/draft-04/schema#';
 use constant VALIDATE_HOSTNAME => eval 'require Data::Validate::Domain;1';
 use constant VALIDATE_IP       => eval 'require Data::Validate::IP;1';
-use constant SPECIFICATION_URL => 'http://json-schema.org/draft-04/schema#';
 
-use constant DEBUG => $ENV{JSON_VALIDATOR_DEBUG} || 0;
-
-our $VERSION = '1.04';
+our $ERR;    # ugly hack to improve validation errors
+our $VERSION = '1.05';
 our @EXPORT_OK = 'validate_json';
 
 my $BUNDLED_CACHE_DIR = path(path(__FILE__)->dirname, qw(Validator cache));
@@ -42,8 +44,7 @@ has cache_paths => sub {
   return \@paths;
 };
 
-has formats  => sub { shift->_build_formats };
-has resolver => sub { \&_resolver };
+has formats => sub { shift->_build_formats };
 
 has ua => sub {
   require Mojo::UserAgent;
@@ -63,33 +64,17 @@ sub coerce {
 
 sub load_and_validate_schema {
   my ($self, $spec, $args) = @_;
-  my $clone = $self->new(%$self)->schema($spec)->schema;
-  my @errors
-    = $self->new(%$self)->schema($args->{schema} || SPECIFICATION_URL)->validate($clone->data);
-
+  $spec = $self->_reset->_resolve($spec);
+  my @errors = $self->new(%$self)->schema($args->{schema} || SPECIFICATION_URL)->validate($spec);
   Carp::confess(join "\n", "Invalid schema:", @errors) if @errors;
-  warn "[JSON::Validator] Loaded $spec\n" if DEBUG;
-  $self->{schema} = $clone;
+  $self->{schema} = Mojo::JSON::Pointer->new($spec);
   $self;
 }
 
 sub schema {
-  my ($self, $schema) = @_;
-
-  if (@_ == 1) {
-    return $self->{schema};
-  }
-  elsif (ref $schema eq 'HASH') {
-    $schema->{id} ||= $self->_default_id($schema);
-    warn "[JSON::Validator] Schema from hash. id=$schema->{id}\n" if DEBUG;
-    $schema = $self->_register_document($schema, $schema->{id});
-  }
-  else {
-    $schema = path($schema)->to_abs if -e $schema;
-    $schema = $self->_load_schema($schema);
-  }
-
-  $self->{schema} = $self->_resolve_schema($schema, $schema->data->{id});
+  my $self = shift;
+  my $schema = shift or return $self->{schema};
+  $self->{schema} = Mojo::JSON::Pointer->new($self->_reset->_resolve($schema));
   $self;
 }
 
@@ -120,59 +105,58 @@ sub _build_formats {
   };
 }
 
-sub _load_schema {
-  my ($self, $url, $parent) = @_;
-  my ($namespace, $scheme) = ("$url", "file");
-  $namespace =~ s|\\|/|g if $^O eq 'MSWin32';
-  my $doc;
+sub _explode {
+  my ($self, $schema) = @_;
+  my @topics = ($schema);
 
-  if ($namespace =~ $HTTP_SCHEME_RE) {
-    $url = Mojo::URL->new($url);
-    ($namespace, $scheme) = ($url->clone->fragment(undef)->to_string, $url->scheme);
-  }
-  elsif ($namespace =~ m!^data://(.*)!) {
-    $scheme = 'data';
-  }
-  elsif ($parent and $parent =~ $HTTP_SCHEME_RE) {
-    $parent = Mojo::URL->new($parent);
-    $url =~ s!#.*!!;
-    $url = $parent->path($parent->path->merge($url)->canonicalize);
-    ($namespace, $scheme) = ($url->to_string, $url->scheme);
-  }
-  elsif ($parent) {
-    $url =~ s!#.*!!;
-    $url = path(path($parent)->dirname, split '/', $url);
-    $namespace = Cwd::abs_path($url->to_string) || $url;
-  }
-  elsif ($url =~ m!^/! and !-e $url and $self->ua->server->app) {
-    $scheme = 'http';
+  while (@topics) {
+    my $topic = shift @topics;
+    if (ref $topic eq 'ARRAY') {
+      push @topics, @$topic;
+    }
+    elsif (ref $topic eq 'HASH') {
+      if ($topic->{'$ref'}) {
+        my $other = $self->_ref_to_schema($topic);
+        %$topic = %$other;
+      }
+      push @topics, values %$topic;
+    }
   }
 
-  # Make sure we create the correct namespace if not already done by Mojo::URL
-  $namespace =~ s!#.*$!! if $namespace eq $url;
-
-  return $doc if $doc = $self->_store(cached => $namespace);
-  return eval {
-    warn "[JSON::Validator] Loading schema $url namespace=$namespace scheme=$scheme\n"
-      if DEBUG;
-    $doc
-      = $scheme eq 'file' ? path($namespace)->slurp
-      : $scheme eq 'data' ? $self->_load_schema_from_data($url, $namespace)
-      :                     $self->_load_schema_from_url($url, $namespace);
-    $self->_register_document($self->_load_schema_from_text(\$doc), $namespace);
-  } || do {
-    $doc ||= '';
-    die "Could not load document from $url: $@ ($doc)" if DEBUG;
-    die "Could not load document from $url: $@";
-  };
+  return $schema;
 }
 
-sub _load_schema_from_data {
-  my ($self, $url, $namespace) = @_;
-  require Mojo::Loader;
-  my ($module, $file) = $namespace =~ m!^data://([^/]+)/(.*)$! or die "Invalid URL: $namespace.";
-  Mojo::Loader::data_section($module, $file)
-    || die "$file could not be found in __DATA__ section of $module.";
+sub _load_schema {
+  my ($self, $url) = @_;
+
+  if ($url =~ m!^https?://!) {
+    warn "[JSON::Validator] Loading schema from URL $url\n" if DEBUG;
+    return $self->_load_schema_from_url(Mojo::URL->new($url)->fragment(undef)), "$url";
+  }
+
+  if ($url =~ m!^data://([^/]+)/(.*)!) {
+    my ($module, $file) = ($1, $2);
+    warn "[JSON::Validator] Loading schema from data section: $url\n" if DEBUG;
+    my $text = Mojo::Loader::data_section($module, $file)
+      || Carp::confess("$file could not be found in __DATA__ section of $module.");
+    return $self->_load_schema_from_text(\$text), "$url";
+  }
+
+  if ($url =~ m!^\s*[\[\{]!) {
+    warn "[JSON::Validator] Loading schema from string.\n" if DEBUG;
+    return $self->_load_schema_from_text(\$url), '';
+  }
+
+  my $file = $url;
+  $file =~ s!#$!!;
+  $file = path(split '/', $file);
+  if (-e $file) {
+    $file = $file->realpath;
+    warn "[JSON::Validator] Loading schema from file: $file\n" if DEBUG;
+    return $self->_load_schema_from_text(\$file->slurp), $url;
+  }
+
+  Carp::confess("Unable to load schema '$url'");
 }
 
 sub _load_schema_from_text {
@@ -196,144 +180,147 @@ sub _load_schema_from_text {
 }
 
 sub _load_schema_from_url {
-  my ($self, $url, $namespace) = @_;
+  my ($self, $url) = @_;
   my $cache_path = $self->cache_paths->[0];
-  my $cache_file = Mojo::Util::md5_sum($namespace);
-  my $tx;
+  my $cache_file = Mojo::Util::md5_sum("$url");
+  my ($err, $tx);
 
   for (@{$self->cache_paths}) {
     my $path = path $_, $cache_file;
     next unless -r $path;
     warn "[JSON::Validator] Loading cached file $path\n" if DEBUG;
-    return $path->slurp;
+    return $self->_load_schema_from_text(\$path->slurp);
   }
 
   $tx = $self->ua->get($url);
-  die $tx->error->{message} if $tx->error;
+  $err = $tx->error && $tx->error->{message};
+  Carp::confess("GET $url == $err") if DEBUG and $err;
+  die "[JSON::Validator] GET $url == $err" if $err;
 
   if ($cache_path and $cache_path ne $BUNDLED_CACHE_DIR and -w $cache_path) {
     $cache_file = path $cache_path, $cache_file;
-    warn "[JSON::Validator] Caching $namespace to $cache_file\n" unless $ENV{HARNESS_ACTIVE};
+    warn "[JSON::Validator] Caching $url to $cache_file\n" unless $ENV{HARNESS_ACTIVE};
     $cache_file->spurt($tx->res->body);
   }
 
-  return $tx->res->body;
+  return $self->_load_schema_from_text(\$tx->res->body);
 }
 
-sub _default_id {
-  state $id = 0;
-  return sprintf 'http://local/json-validator/default-id-%s', ++$id;
-}
+sub _ref_to_schema {
+  my ($self, $schema) = @_;
 
-sub _register_document {
-  my ($self, $doc, $namespace) = @_;
-
-  $doc = Mojo::JSON::Pointer->new($doc);
-  $namespace = Mojo::URL->new($namespace) unless ref $namespace;
-  $namespace->fragment(undef);
-  $doc->data->{id} ||= "$namespace";
-  $self->_store(cached => $namespace       => $doc);
-  $self->_store(cached => $doc->data->{id} => $doc);
-
-  warn "[JSON::Validator] Register id=$doc->{data}{id} namespace=$namespace\n" if DEBUG;
-  return $doc;
-}
-
-# This method is used to resolve all the $ref's that we find inside JSON Schema
-# specification.
-#
-# $namespace is typically the "id" part at the top level of the schema
-# specification. Meaning some kind of URL. In the case of Swagger: Just make
-# something up instead using _default_id()
-sub _resolve_schema {
-  my ($self, $schema, $namespace) = @_;
-  my ($resolved, @topics, @refs);
-
-  # The if is true if we have already resolved this $namespace. This happens if
-  # the same specification has "$ref" with the same value multiple times.
-  # There's no need to resolve the same schema twice.
-  return $resolved if $resolved = $self->_store(resolved => $namespace);
-
-  @topics = ($self->_store(resolved => $namespace, $schema))->data;
-
-  # This while loop will traverse the whole specification and and track down
-  # each and every "$ref" it finds. @topics is a list of all the data
-  # structures that it finds. Note that is start out with just having one item:
-  # A hash-ref to the complete specification, but as it goes along, the while()
-  # loop will track down more objects and arrays inside the specification and
-  # then loop over those as well, util it has visited the whole document.
-  while (@topics) {
-    my $topic = shift @topics;
-    next if ref $topic and $self->_seen(topic => $topic);    # Avoid recursion
-    if (UNIVERSAL::isa($topic, 'HASH')) {
-      for my $k (sort keys %$topic) {
-        my $v = $topic->{$k};
-        push @topics, $topic->{$k} if ref $topic->{$k};
-        unshift @refs, $topic if $k eq '$ref' and !ref $v;
-      }
-    }
-    elsif (UNIVERSAL::isa($topic, 'ARRAY')) {
-      unshift @topics, @$topic;
-    }
+  my @guard;
+  while ($schema->{'$ref'}) {
+    push @guard, $schema->{'$ref'};
+    Carp::confess("Seems like you have a circular reference: @guard") if @guard > RECURSION_LIMIT;
+    $schema = $self->{refs}{refaddr($schema)} // Carp::confess("Could not lookup @guard");
   }
 
-  # Resolve all the "$ref" we found. This will call _resolve_schema() again.
-  $self->resolver->($self, $namespace, \@refs);
-
   return $schema;
 }
 
-sub _resolver {
-  my ($self, $namespace, $refs) = @_;
+sub _register_ref {
+  my ($self, $topic, $schema_url) = @_;
+  my $ref = $topic->{'$ref'};
+  my $fqn = Mojo::URL->new($ref =~ m!^/! ? "#$ref" : $ref);
 
-  # Seconds step: Resolve $ref
-  for my $topic (@$refs) {
-    my $ref = delete $topic->{'$ref'} or next;    # already resolved?
-    $ref = "#/definitions/$ref" if $ref =~ /^\w+$/;    # TODO: Figure out if this could be removed
-    $ref = Mojo::Util::url_unescape($ref || '');
-    $ref = Mojo::URL->new($namespace)->fragment($ref) if $ref =~ s!^\#!!;
-    $ref = Mojo::URL->new($ref) unless ref $ref;
+  $fqn = $fqn->to_abs($schema_url) unless $fqn->is_abs;
+  $fqn = $fqn->to_string;
+  $fqn =~ s!(.)#$!$1!;
+  $fqn =~ s!#(.+)!{'#' . url_unescape $1}!e;
 
-    my $look_in = $self->_store(resolved => $ref);
-    if (!$look_in) {
-      $look_in = $self->_load_schema($ref, $namespace);
-      $look_in = $self->_resolve_schema($look_in, $look_in->data->{id} || $namespace);
-    }
+  my ($base, $pointer) = split /#/, $fqn, 2;
+  my $other = $self->_resolve($base);
 
-    warn "[JSON::Validator] Resolving $ref\n" if DEBUG > 1;
-    warn Data::Dumper::Dumper($look_in->data) if DEBUG and $ref =~ /\b[c]\b/;   # follow the changes
-    my $data = $ref->fragment ? $look_in->get($ref->fragment) : $look_in->data;
-    die qq[Possibly a typo in schema? Could not find "$ref"] unless $data;
-    $topic->{$_} = $data->{$_} for keys %$data;
-    unshift @$refs, $topic if $topic->{'$ref'} and !ref $topic->{'$ref'};
-    delete $topic->{id} if !ref $topic->{id} and $self->isa('JSON::Validator::OpenAPI');
+  if (length $pointer) {
+    $other = Mojo::JSON::Pointer->new($other)->get($pointer)
+      or Carp::confess(qq[Possibly a typo in schema? Could not find "$ref" ($fqn)]);
   }
+
+  $self->{refs}{refaddr($topic)} = $other;
 }
 
-# this code is from Data::Dumper::format_refaddr()
+sub _register_schema {
+  my ($self, $schema, $fqn) = @_;
+  $fqn =~ s!(.)#$!$1!;
+  $self->{schemas}{$fqn} = $schema;
+}
+
+sub _reset {
+  delete $_[0]->{$_} for qw(refs schemas);
+  $_[0]->{level} = 0;
+  $_[0];
+}
+
+# _resolve() method is used to convert all "id" into absolute URLs and
+# resolve all the $ref's that we find inside JSON Schema specification.
+sub _resolve {
+  my ($self, $schema) = @_;
+  my ($id, $resolved, @refs);
+
+  if (ref $schema eq 'HASH') {
+    $id = $schema->{id} // '';
+    return $resolved if $resolved = $self->{schemas}{$id};
+  }
+  elsif ($resolved = $self->{schemas}{$schema // ''}) {
+    return $resolved;
+  }
+  else {
+    ($schema, $id) = $self->_load_schema($schema);
+    $id = $schema->{id} if $schema->{id};
+  }
+
+  if (!$self->{level}++ and my $id = $schema->{id}) {
+    Carp::confess("Root schema cannot have a fragment in the 'id'. ($id)") if $id =~ /\#./;
+    Carp::confess("Root schema cannot have a relative 'id'. ($id)") if $id and $id !~ /^\w+:/;
+  }
+
+  $self->_register_schema($schema, $id);
+
+  my @topics = ([$schema, Mojo::URL->new($id)]);
+  while (@topics) {
+    my ($topic, $base) = @{shift @topics};
+
+    if (UNIVERSAL::isa($topic, 'ARRAY')) {
+      push @topics, map { [$_, $base] } @$topic;
+    }
+    elsif (UNIVERSAL::isa($topic, 'HASH')) {
+      if ($topic->{'$ref'} and !ref $topic->{'$ref'}) {
+        push @refs, [$topic, $base];
+      }
+      if ($topic->{id} and !ref $topic->{id}) {
+        my $fqn = Mojo::URL->new($topic->{id});
+        $fqn = $fqn->to_abs($base) unless $fqn->is_abs;
+        $self->_register_schema($topic, $fqn->to_string);
+      }
+
+      push @topics, map { [$topic->{$_}, $base] } grep { $_ ne '$ref' } keys %$topic;
+    }
+  }
+
+  $self->_register_ref(@$_) for @refs;
+
+  return $schema;
+}
+
+# This code is from Data::Dumper::format_refaddr()
 sub _seen {
-  require Scalar::Util;
   my $self = shift;
-  my $key = join ':', shift, map { pack 'J', Scalar::Util::refaddr($_) } @_;
+  my $key = join ':', shift, map { pack 'J', refaddr $_ } @_;
   return $self->{seen}{$key}++;
 }
 
-sub _store {
-  my ($self, $key, $namespace, $schema) = @_;
-  $namespace = Mojo::URL->new($namespace)->fragment(undef)->to_string;
-  return $self->{$key}{$namespace} unless $schema;
-  return $self->{$key}{$namespace} = $schema;
-}
-
 sub _validate {
   my ($self, $data, $path, $schema) = @_;
   my ($type, @errors);
 
+  $schema = $self->_ref_to_schema($schema) if $schema->{'$ref'};
+
   # Avoid recursion
-  return if ref $data and $self->_seen(data => $schema, $data);
+  return if ref $data and !_is_blessed_boolean($data) and $self->_seen($schema, $data);
 
   # Make sure we validate plain data and not a perl object
-  $data = $data->TO_JSON if Scalar::Util::blessed($data) and UNIVERSAL::can($data, 'TO_JSON');
+  $data = $data->TO_JSON if blessed $data and UNIVERSAL::can($data, 'TO_JSON');
   $type = $schema->{type} || _guess_schema_type($schema, $data);
 
   # Test base schema before allOf, anyOf or oneOf
@@ -471,9 +458,10 @@ sub _validate_type_const {
 sub _validate_format {
   my ($self, $value, $path, $schema) = @_;
   my $code = $self->formats->{$schema->{format}};
+  local $ERR;
   return if $code and $code->($value);
   return do { warn "Format rule for '$schema->{format}' is missing"; return } unless $code;
-  return E $path, "Does not match $schema->{format} format.";
+  return E $path, $ERR || "Does not match $schema->{format} format.";
 }
 
 sub _validate_type_any { }
@@ -528,8 +516,7 @@ sub _validate_type_array {
 sub _validate_type_boolean {
   my ($self, $value, $path, $schema) = @_;
 
-  return if UNIVERSAL::isa($value, 'JSON::PP::Boolean');
-  return if Scalar::Util::blessed($value) and ("$value" eq "1" or !$value);
+  return if _is_blessed_boolean($value);
 
   if (  defined $value
     and $self->{coerce}{booleans}
@@ -719,7 +706,7 @@ sub _expected {
 sub _guess_data_type {
   local $_ = $_[0];
   my $ref     = ref;
-  my $blessed = Scalar::Util::blessed($_[0]);
+  my $blessed = blessed $_;
   return 'object' if $ref eq 'HASH';
   return lc $ref if $ref and !$blessed;
   return 'null' if !defined;
@@ -759,6 +746,7 @@ sub _guessed_right {
 }
 
 sub _invalid {
+  $ERR = $_[0];
   warn sprintf "[JSON::Validator] Failed validation: $_[0]\n" if DEBUG;
   return 0;
 }
@@ -799,38 +787,41 @@ sub _is_ipv6 { warn "Data::Validate::IP is not installed"; return; }
 
 sub _is_true {
   local $_ = $_[0];
-  return 0 + $_ if ref $_ and !Scalar::Util::blessed($_);
+  return 0 + $_ if ref $_ and !blessed $_;
   return 0 if !$_ or /^(n|false|off)/i;
   return 1;
 }
 
+sub _is_blessed_boolean {
+  return 0 if !blessed $_[0];
+  return 1 if UNIVERSAL::isa($_[0], 'JSON::PP::Boolean') or "$_[0]" eq "1" or !$_[0];
+  return 0;
+}
+
 sub _is_regex {
   eval {qr{$_[0]}};
 }
 
-# From Data::Validate::URI
 sub _is_uri {
-  return unless $_[0];
+  return unless defined $_[0];
+  return unless $_[0] =~ m!^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?!;
 
-  my ($scheme, $authority, $path, $query, $fragment)
-    = $_[0] =~ qr!^(?:([^:/?#]+):)?(?://([^/?#]*))?([^?#]*)(?:\?([^#]*))?(?:#(.*))?$!o;
+  my ($scheme, $auth_host, $path, $query, $fragment) = map { $_ // '' } ($2, $4, $5, $7, $9);
 
-  return _invalid('Scheme and path are required')
-    unless defined $scheme
-    and length $scheme
-    and defined $path;
+  return _invalid('Scheme missing from URI.') if length $auth_host and !length $scheme;
+  return _invalid('Scheme, path or fragment are required.')
+    unless length($scheme) + length($path) + length($fragment);
+  return _invalid('Scheme must begin with a letter.')
+    if length $scheme and lc($scheme) !~ m!^[a-z][a-z0-9\+\-\.]*$!;
+  return _invalid('Invalid hex escape.')           if $_[0] =~ /%[^0-9a-f]/i;
+  return _invalid('Hex escapes are not complete.') if $_[0] =~ /%[0-9a-f](:?[^0-9a-f]|$)/i;
 
-  return _invalid('Scheme must begin with a letter')
-    unless lc($scheme) =~ m!^[a-z][a-z0-9\+\-\.]*$!;
-
-  return _invalid('Hex escapes are not complete') if $_[0] =~ /%[^0-9a-f]/i;
-  return _invalid('Hex escapes are not complete') if $_[0] =~ /%[0-9a-f](:?[^0-9a-f]|$)/i;
-
-  if (defined $authority and length $authority) {
-    return _invalid('Path must be empty or begin with a /') unless !length $path or $path =~ m!^/!;
+  if (defined $auth_host and length $auth_host) {
+    return _invalid('Path cannot be empty or begin with a /')
+      unless !length $path or $path =~ m!^/!;
   }
   else {
-    return _invalid('Path must not start with //') if $path =~ m!^//!;
+    return _invalid('Path cannot not start with //.') if $path =~ m!^//!;
   }
 
   return 1;
@@ -879,7 +870,7 @@ JSON::Validator - Validate data against a JSON schema
 
 =head1 VERSION
 
-1.04
+1.05
 
 =head1 SYNOPSIS
 
@@ -1120,18 +1111,6 @@ Validated against the RFC3986 spec.
 
 =back
 
-=head2 resolver
-
-  $code = $self->resolver;
-  $self = $self->resolver(sub { my ($self, $namespace, $refs) = @_; });
-
-Set this to a sub without any logic if you want to skip resolving references,
-like this:
-
-  $self->resolver(sub {});
-
-This attribute is EXPERIMENTAL.
-
 =head2 ua
 
   $ua = $self->ua;
